@@ -20,7 +20,6 @@
 #ifndef BDJ_PRIVATE_H_
 #define BDJ_PRIVATE_H_
 
-#include "libbluray/register.h"
 #include "libbluray/bluray.h"
 #include "libbluray/bdnav/index_parse.h"
 #include "libbluray/decoders/overlay.h"
@@ -28,7 +27,6 @@
 
 struct bdjava_s {
     BLURAY       *bd;
-    BD_REGISTERS *reg;
     INDX_ROOT    *index;
 
     bdj_overlay_cb  osd_cb;
@@ -41,6 +39,8 @@ struct bdjava_s {
     JavaVM* jvm;
 
     const char *path;
+
+    unsigned uo_mask; /* UO masks from bdjo */
 };
 
 #endif
