@@ -136,7 +136,7 @@ public class BDWindowGraphics extends BDGraphics {
         }
     }
 
-    public void drawStringN(long ftFace, String string, int x, int y, int rgb) {
+    protected void drawStringN(long ftFace, String string, int x, int y, int rgb) {
         synchronized (window) {
             super.drawStringN(ftFace, string, x, y, rgb);
             window.notifyChanged();
@@ -151,11 +151,13 @@ public class BDWindowGraphics extends BDGraphics {
     public boolean drawImageN(Image img,
         int dx, int dy, int dw, int dh,
         int sx, int sy, int sw, int sh,
+        boolean flipX, boolean flipY,
         Color bg, ImageObserver observer) {
 
         synchronized (window) {
             boolean complete = super.drawImageN(
                 img, dx, dy, dw, dh, sx, sy, sw, sh,
+                flipX, flipY,
                 bg, observer);
             if (complete) {
                 window.notifyChanged();
