@@ -40,7 +40,9 @@ import javax.media.NotPrefetchedError;
 import javax.media.NotRealizedError;
 import javax.media.Player;
 import javax.media.PrefetchCompleteEvent;
+import javax.media.RateChangeEvent;
 import javax.media.RealizeCompleteEvent;
+import javax.media.ResourceUnavailableEvent;
 import javax.media.StartEvent;
 import javax.media.StopByRequestEvent;
 import javax.media.Time;
@@ -56,12 +58,28 @@ import org.bluray.net.BDLocator;
 
 import org.videolan.BDJAction;
 import org.videolan.BDJActionManager;
+import org.videolan.BDJActionQueue;
 import org.videolan.BDJListeners;
 import org.videolan.BDJXletContext;
 import org.videolan.Logger;
 
 public abstract class BDHandler implements Player, ServiceContentHandler {
+
     public BDHandler() {
+        ownerContext = BDJXletContext.getCurrentContext();
+
+        PlayerAction action = new PlayerAction(this, PlayerAction.ACTION_INIT, null);
+        BDJActionManager.getInstance().putCommand(action);
+        action.waitEnd();
+    }
+
+    private void doInitAction() {
+        commandQueue = new BDJActionQueue("MediaPlayer");
+        PlayerManager.getInstance().registerPlayer(this);
+    }
+
+    protected BDJXletContext getOwnerContext() {
+        return ownerContext;
     }
 
     private void checkUnrealized() {
@@ -187,8 +205,11 @@ public abstract class BDHandler implements Player, ServiceContentHandler {
 
     public void setMediaTime(Time now) {
         checkUnrealized();
+
+        if (isClosed) return;
+
         PlayerAction action = new PlayerAction(this, PlayerAction.ACTION_SEEK_TIME, now);
-        BDJActionManager.getInstance().putCommand(action);
+        commandQueue.put(action);
         action.waitEnd();
     }
 
@@ -205,14 +226,23 @@ public abstract class BDHandler implements Player, ServiceContentHandler {
         baseTime = getTimeBase().getNanoseconds();
     }
 
+    /* notification from app */
+    protected void updateRate(float rate) {
+        if (this.rate != rate) {
+            this.rate = rate;
+            notifyListeners(new RateChangeEvent(this, rate));
+        }
+    }
+
     public float getRate() {
         return rate;
     }
 
     public float setRate(float factor) {
         checkUnrealized();
+
         PlayerAction action = new PlayerAction(this, PlayerAction.ACTION_SET_RATE, new Float(factor));
-        BDJActionManager.getInstance().putCommand(action);
+        commandQueue.put(action);
         action.waitEnd();
         return rate;
     }
@@ -226,13 +256,17 @@ public abstract class BDHandler implements Player, ServiceContentHandler {
     }
 
     public void realize() {
+        if (isClosed) return;
+
         PlayerAction action = new PlayerAction(this, PlayerAction.ACTION_REALIZE, null);
-        BDJActionManager.getInstance().putCommand(action);
+        commandQueue.put(action);
     }
 
     public void prefetch() {
+        if (isClosed) return;
+
         PlayerAction action = new PlayerAction(this, PlayerAction.ACTION_PREFETCH, null);
-        BDJActionManager.getInstance().putCommand(action);
+        commandQueue.put(action);
     }
 
     public void syncStart(Time at) {
@@ -240,36 +274,58 @@ public abstract class BDHandler implements Player, ServiceContentHandler {
             if (state != Prefetched)
                 throw new NotPrefetchedError("syncStart");
         }
+        if (isClosed) return;
+
         PlayerAction action = new PlayerAction(this, PlayerAction.ACTION_START, at);
-        BDJActionManager.getInstance().putCommand(action);
+        commandQueue.put(action);
     }
 
     public void start() {
+        if (isClosed) return;
+
         PlayerAction action = new PlayerAction(this, PlayerAction.ACTION_START, null);
-        BDJActionManager.getInstance().putCommand(action);
+        commandQueue.put(action);
     }
 
     public void stop() {
+        if (isClosed) return;
+
         PlayerAction action = new PlayerAction(this, PlayerAction.ACTION_STOP, null);
-        BDJActionManager.getInstance().putCommand(action);
+        commandQueue.put(action);
         action.waitEnd();
     }
 
     public void deallocate() {
+        if (isClosed) return;
+
+        if (state == Started) {
+        }
         PlayerAction action = new PlayerAction(this, PlayerAction.ACTION_DEALLOCATE, null);
-        BDJActionManager.getInstance().putCommand(action);
+        commandQueue.put(action);
         action.waitEnd();
+
+        PlayerManager.getInstance().releaseResource(this);
     }
 
     public void close() {
+        if (isClosed) return;
+
+        stop();
+        deallocate();
+
         PlayerAction action = new PlayerAction(this, PlayerAction.ACTION_CLOSE, null);
-        BDJActionManager.getInstance().putCommand(action);
+        commandQueue.put(action);
         action.waitEnd();
+
+        isClosed = true;
+        commandQueue.shutdown();
     }
 
-    protected void endOfMedia() {
+    protected void endOfMedia(int playlist) {
+        if (isClosed) return;
+
         PlayerAction action = new PlayerAction(this, PlayerAction.ACTION_END_OF_MEDIA, null);
-        BDJActionManager.getInstance().putCommand(action);
+        commandQueue.put(action);
     }
 
     protected void updateTime(int time) {
@@ -277,68 +333,13 @@ public abstract class BDHandler implements Player, ServiceContentHandler {
         //currentTime = new Time(time * TO_SECONDS);
     }
 
-    public static void activePlayerEndOfMedia() {
-        synchronized (BDHandler.class) {
-            if (activePlayer != null)
-                activePlayer.endOfMedia();
-        }
-    }
-
-    public static void activePlayerUpdateTime(int pts) {
-        synchronized (BDHandler.class) {
-            if (activePlayer != null)
-                activePlayer.updateTime(pts);
-        }
-    }
-
-    public static void onChapterReach(int param) {
-        synchronized (BDHandler.class) {
-            if (activePlayer != null)
-                activePlayer.doChapterReach(param);
-        }
-    }
-
-    public static void onMarkReach(int param) {
-        synchronized (BDHandler.class) {
-            if (activePlayer != null)
-                activePlayer.doMarkReach(param);
-        }
-    }
-
-    public static void onPlayItemReach(int param) {
-        synchronized (BDHandler.class) {
-            if (activePlayer != null)
-                activePlayer.doPlayItemReach(param);
-        }
-    }
-
-    public static void onAngleChange(int param) {
-        synchronized (BDHandler.class) {
-            if (activePlayer != null)
-                activePlayer.doAngleChange(param);
-        }
-    }
-
-    public static void onSubtitleChange(int param) {
-        synchronized (BDHandler.class) {
-            if (activePlayer != null)
-                activePlayer.doSubtitleChange(param);
-        }
-    }
-
-    public static void onPiPChange(int param) {
-        synchronized (BDHandler.class) {
-            if (activePlayer != null)
-                activePlayer.doPiPChange(param);
-        }
-    }
-
-    protected abstract void doChapterReach(int param);
-    protected abstract void doMarkReach(int param);
-    protected abstract void doPlayItemReach(int param);
-    protected abstract void doAngleChange(int param);
-    protected abstract void doSubtitleChange(int param);
-    protected abstract void doPiPChange(int param);
+    protected void doPlaylistStart(int param) {};
+    protected void doChapterReach(int param) {};
+    protected void doMarkReach(int param) {};
+    protected void doPlayItemReach(int param) {};
+    protected void doAngleChange(int param) {};
+    protected void doSubtitleChange(int param) {};
+    protected void doPiPChange(int param) {};
 
     protected ControllerErrorEvent doRealize() {
         return null;
@@ -373,7 +374,10 @@ public abstract class BDHandler implements Player, ServiceContentHandler {
     }
 
     protected void doSetRate(Float factor) {
-        rate = factor.floatValue();
+        if (rate != factor.floatValue()) {
+            rate = factor.floatValue();
+            notifyListeners(new RateChangeEvent(this, rate));
+        }
     }
 
     private void notifyListeners(ControllerEvent event) {
@@ -412,10 +416,10 @@ public abstract class BDHandler implements Player, ServiceContentHandler {
             state = Prefetching;
             notifyListeners(new TransitionEvent(this, Realized, Prefetching, Prefetched));
         case Prefetching:
-            synchronized (BDHandler.class) {
-                if (activePlayer != null)
-                    activePlayer.doDeallocateAction();
-                activePlayer = this;
+
+            if (!PlayerManager.getInstance().allocateResource(this)) {
+                notifyListeners(new ResourceUnavailableEvent(this));
+                return false;
             }
             ControllerErrorEvent error = doPrefetch();
             if (error == null) {
@@ -499,10 +503,6 @@ public abstract class BDHandler implements Player, ServiceContentHandler {
         default:
             error = doDeallocate();
             if (error == null) {
-                synchronized (BDHandler.class) {
-                    if (activePlayer == this)
-                        activePlayer = null;
-                }
                 int previous = state;
                 state = Realized;
                 notifyListeners(new DeallocateEvent(this, previous, Realized, Realized, getMediaTime()));
@@ -518,6 +518,7 @@ public abstract class BDHandler implements Player, ServiceContentHandler {
     private void doCloseAction() {
         doClose();
         notifyListeners(new ControllerClosedEvent(this));
+        PlayerManager.getInstance().unregisterPlayer(this);
     }
 
     private void doEndOfMediaAction() {
@@ -541,6 +542,9 @@ public abstract class BDHandler implements Player, ServiceContentHandler {
 
         protected void doAction() {
             switch (action) {
+            case ACTION_INIT:
+                player.doInitAction();
+                break;
             case ACTION_REALIZE:
                 player.doRealizeAction();
                 break;
@@ -584,6 +588,7 @@ public abstract class BDHandler implements Player, ServiceContentHandler {
         public static final int ACTION_END_OF_MEDIA = 7;
         public static final int ACTION_SEEK_TIME = 8;
         public static final int ACTION_SET_RATE = 9;
+        public static final int ACTION_INIT = 10;
     }
 
     protected int state = Unrealized;
@@ -595,10 +600,13 @@ public abstract class BDHandler implements Player, ServiceContentHandler {
     protected Control[] controls = null;
     protected BDLocator locator = null;
     private BDJListeners listeners = new BDJListeners();
+    private BDJXletContext ownerContext;
+    boolean isClosed = false;
+
+    protected BDJActionQueue commandQueue;
 
     public static final double TO_SECONDS = 1 / 90000.0d;
     public static final double FROM_SECONDS = 90000.0d;
     public static final double TO_NAROSECONDS = 1000000 / 90.0d;
     public static final double FROM_NAROSECONDS = 0.00009d;
-    public static BDHandler activePlayer = null;
 }
