@@ -46,7 +46,7 @@ int bdj_register_native_methods(JNIEnv *env)
                            Java_java_awt_BDFontMetrics_methods,
                            Java_java_awt_BDFontMetrics_methods_count)
       *
-      bdj_register_methods(env, "java/awt/BDGraphics",
+      bdj_register_methods(env, "java/awt/BDGraphicsBase",
                            Java_java_awt_BDGraphics_methods,
                            Java_java_awt_BDGraphics_methods_count)
       ;
@@ -54,7 +54,7 @@ int bdj_register_native_methods(JNIEnv *env)
 
 void bdj_unregister_native_methods(JNIEnv *env)
 {
-    bdj_unregister_methods(env, "java/awt/BDGraphics");
+    bdj_unregister_methods(env, "java/awt/BDGraphicsBase");
     bdj_unregister_methods(env, "java/awt/BDFontMetrics");
     bdj_unregister_methods(env, "org/videolan/Libbluray");
     bdj_unregister_methods(env, "org/videolan/Logger");
