@@ -44,6 +44,7 @@
 #endif
 
 #define BAD_CAST_CONST (const xmlChar *)
+#define XML_FREE(p) (xmlFree(p), p = NULL)
 
 #define MAX_META_FILE_SIZE  0xfffff
 
@@ -64,11 +65,11 @@ static void _parseManifestNode(xmlNode * a_node, META_DL *disclib)
                 }
                 if (xmlStrEqual(cur_node->name, BAD_CAST_CONST "numSets")) {
                     disclib->di_num_sets = atoi((char*)(tmp = xmlNodeGetContent(cur_node)));
-                    xmlFree(tmp);
+                    XML_FREE(tmp);
                 }
                 if (xmlStrEqual(cur_node->name, BAD_CAST_CONST "setNumber")) {
                     disclib->di_set_number = atoi((char*)(tmp = xmlNodeGetContent(cur_node)));
-                    xmlFree(tmp);
+                    XML_FREE(tmp);
                 }
             }
             else if (xmlStrEqual(cur_node->parent->name, BAD_CAST_CONST "tableOfContents")) {
@@ -78,7 +79,7 @@ static void _parseManifestNode(xmlNode * a_node, META_DL *disclib)
                     disclib->toc_entries = realloc(disclib->toc_entries, (disclib->toc_count*sizeof(META_TITLE)));
                     disclib->toc_entries[i].title_number = atoi((const char*)tmp);
                     disclib->toc_entries[i].title_name = (char*)xmlNodeGetContent(cur_node);
-                    X_FREE(tmp);
+                    XML_FREE(tmp);
                 }
             }
             else if (xmlStrEqual(cur_node->parent->name, BAD_CAST_CONST "description")) {
@@ -92,7 +93,7 @@ static void _parseManifestNode(xmlNode * a_node, META_DL *disclib)
                         sscanf((const char*)tmp, "%ix%i", &x, &y);
                         disclib->thumbnails[i].xres = x;
                         disclib->thumbnails[i].yres = y;
-                        X_FREE(tmp);
+                        XML_FREE(tmp);
                     }
                     else {
                         disclib->thumbnails[i].xres = disclib->thumbnails[i].yres = -1;
@@ -124,6 +125,8 @@ static void _findMetaXMLfiles(META_ROOT *meta, const char *device_path)
             uint8_t i = meta->dl_count;
             meta->dl_count++;
             meta->dl_entries = realloc(meta->dl_entries, (meta->dl_count*sizeof(META_DL)));
+            memset(&meta->dl_entries[i], 0, sizeof(meta->dl_entries[i]));
+
             meta->dl_entries[i].filename = str_dup(ent.d_name);
             strncpy(meta->dl_entries[i].language_code, ent.d_name+5,3);
             meta->dl_entries[i].language_code[3] = '\0';
@@ -180,7 +183,7 @@ META_ROOT *meta_parse(const char *device_path)
                     root->dl_entries[i].toc_entries = NULL;
                     root->dl_entries[i].thumbnails = NULL;
                     _parseManifestNode(root_element, &root->dl_entries[i]);
-                    xmlFreeDoc(doc);
+                    XML_FREE(doc);
                 }
             }
             X_FREE(data);
@@ -199,6 +202,7 @@ META_ROOT *meta_parse(const char *device_path)
 
 const META_DL *meta_get(const META_ROOT *meta_root, const char *language_code)
 {
+#ifdef HAVE_LIBXML2
     unsigned i;
 
     if (meta_root == NULL || meta_root->dl_count == 0) {
@@ -224,28 +228,33 @@ const META_DL *meta_get(const META_ROOT *meta_root, const char *language_code)
 
     BD_DEBUG(DBG_DIR, "requested disclib language '%s' or default '"DEFAULT_LANGUAGE"' not found, using '%s' instead\n", language_code, meta_root->dl_entries[0].language_code);
     return &meta_root->dl_entries[0];
+#else
+    return NULL;
+#endif
 }
 
 void meta_free(META_ROOT **p)
 {
+#ifdef HAVE_LIBXML2
     if (p && *p)
     {
         uint8_t i;
         for (i = 0; i < (*p)->dl_count; i++) {
             uint32_t t;
             for (t = 0; t < (*p)->dl_entries[i].toc_count; t++) {
-                X_FREE((*p)->dl_entries[i].toc_entries[t].title_name);
+                XML_FREE((*p)->dl_entries[i].toc_entries[t].title_name);
             }
             for (t = 0; t < (*p)->dl_entries[i].thumb_count; t++) {
-                X_FREE((*p)->dl_entries[i].thumbnails[t].path);
+                XML_FREE((*p)->dl_entries[i].thumbnails[t].path);
             }
             X_FREE((*p)->dl_entries[i].toc_entries);
             X_FREE((*p)->dl_entries[i].thumbnails);
             X_FREE((*p)->dl_entries[i].filename);
-            X_FREE((*p)->dl_entries[i].di_name);
-            X_FREE((*p)->dl_entries[i].di_alternative);
+            XML_FREE((*p)->dl_entries[i].di_name);
+            XML_FREE((*p)->dl_entries[i].di_alternative);
         }
         X_FREE((*p)->dl_entries);
         X_FREE(*p);
     }
+#endif
 }
