@@ -27,6 +27,7 @@
 #include "bluray.h"
 #include "bluray_internal.h"
 #include "register.h"
+#include "util/array.h"
 #include "util/macro.h"
 #include "util/logging.h"
 #include "util/strutl.h"
@@ -116,7 +117,7 @@ struct bluray {
     /* current disc */
     char             *device_path;
     BLURAY_DISC_INFO  disc_info;
-    INDX_ROOT        *index;
+    BLURAY_TITLE    **titles;  /* titles from disc index */
     META_ROOT        *meta;
     NAV_TITLE_LIST   *title_list;
 
@@ -279,7 +280,6 @@ static void _update_stream_psr_by_lang(BD_REGISTERS *regs,
                                        MPLS_STREAM *streams, unsigned num_streams,
                                        uint32_t *lang, uint32_t blacklist)
 {
-    uint32_t psr_val;
     uint32_t preferred_lang;
     int      stream_idx = -1;
     unsigned ii;
@@ -321,13 +321,9 @@ static void _update_stream_psr_by_lang(BD_REGISTERS *regs,
 
     BD_DEBUG(DBG_BLURAY, "Selected stream %d (language %s)\n", stream_idx, streams[stream_idx].lang);
 
-    bd_psr_lock(regs);
-
-    psr_val = bd_psr_read(regs, psr_stream) & 0x7fff0000;
-    psr_val |= (stream_idx + 1) | enable_flag;
-    bd_psr_write(regs, psr_stream, psr_val);
-
-    bd_psr_unlock(regs);
+    bd_psr_write_bits(regs, psr_stream,
+                      (stream_idx + 1) | enable_flag,
+                      0x80000fff);
 }
 
 static void _update_clip_psrs(BLURAY *bd, NAV_CLIP *clip)
@@ -357,11 +353,12 @@ static void _update_clip_psrs(BLURAY *bd, NAV_CLIP *clip)
                                        NULL, audio_lang);
         }
 
-    /* Validate selected audio and subtitle stream PSRs when using menus */
+    /* Validate selected audio, subtitle and IG stream PSRs when using menus */
     } else {
         uint32_t psr_val;
 
         if (stn->num_audio) {
+            bd_psr_lock(bd->regs);
             psr_val = bd_psr_read(bd->regs, PSR_PRIMARY_AUDIO_ID);
             if (psr_val == 0 || psr_val > stn->num_audio) {
                 _update_stream_psr_by_lang(bd->regs,
@@ -371,8 +368,10 @@ static void _update_clip_psrs(BLURAY *bd, NAV_CLIP *clip)
             } else {
                 audio_lang = str_to_uint32((const char *)stn->audio[psr_val - 1].lang, 3);
             }
+            bd_psr_unlock(bd->regs);
         }
         if (stn->num_pg) {
+            bd_psr_lock(bd->regs);
             psr_val = bd_psr_read(bd->regs, PSR_PG_STREAM) & 0xfff;
             if ((psr_val == 0) || (psr_val > stn->num_pg)) {
                 _update_stream_psr_by_lang(bd->regs,
@@ -380,14 +379,40 @@ static void _update_clip_psrs(BLURAY *bd, NAV_CLIP *clip)
                                            stn->pg, stn->num_pg,
                                            NULL, audio_lang);
             }
+            bd_psr_unlock(bd->regs);
+        }
+        if (stn->num_ig) {
+            bd_psr_lock(bd->regs);
+            psr_val = bd_psr_read(bd->regs, PSR_IG_STREAM_ID);
+            if ((psr_val == 0) || (psr_val > stn->num_ig)) {
+                bd_psr_write(bd->regs, PSR_IG_STREAM_ID, 1);
+                BD_DEBUG(DBG_BLURAY | DBG_CRIT, "Selected IG stream 1 (stream %d not available)\n", psr_val);
+            }
+            bd_psr_unlock(bd->regs);
         }
     }
 }
 
+static int _is_interactive_title(BLURAY *bd)
+{
+    if (bd->title && bd->title_type != title_undef) {
+        unsigned title = bd_psr_read(bd->regs, PSR_TITLE_NUMBER);
+        if (title == 0xffff && bd->disc_info.first_play->interactive) {
+            return 1;
+        }
+        if (title <= bd->disc_info.num_titles && bd->titles[title]) {
+            return bd->titles[title]->interactive;
+        }
+    }
+    return 0;
+}
+
 static void _update_chapter_psr(BLURAY *bd)
 {
-    uint32_t current_chapter = bd_get_current_chapter(bd);
-    bd_psr_write(bd->regs, PSR_CHAPTER,  current_chapter + 1);
+    if (!_is_interactive_title(bd)) {
+        uint32_t current_chapter = bd_get_current_chapter(bd);
+        bd_psr_write(bd->regs, PSR_CHAPTER,  current_chapter + 1);
+    }
 }
 
 /*
@@ -396,7 +421,8 @@ static void _update_chapter_psr(BLURAY *bd)
 
 static int _find_pg_stream(BLURAY *bd, uint16_t *pid, int *sub_path_idx, unsigned *sub_clip_idx, uint8_t *char_code)
 {
-    MPLS_PI  *pi        = &bd->title->pl->play_item[0];
+    unsigned  main_clip_idx = bd->st0.clip ? bd->st0.clip->ref : 0;
+    MPLS_PI  *pi        = &bd->title->pl->play_item[main_clip_idx];
     unsigned  pg_stream = bd_psr_read(bd->regs, PSR_PG_STREAM);
 
 #if 0
@@ -464,7 +490,7 @@ static void _update_textst_timer(BLURAY *bd)
 {
     if (bd->st_textst.clip) {
         if (bd->st0.clip_block_pos >= bd->gc_wakeup_pos) {
-            GC_NAV_CMDS cmds = {-1, NULL, -1, 0, 0};
+            GC_NAV_CMDS cmds = {-1, NULL, -1, 0, 0, EMPTY_UO_MASK};
 
             gc_run(bd->graphics_controller, GC_CTRL_PG_UPDATE, bd->gc_wakeup_time, &cmds);
 
@@ -761,7 +787,7 @@ static int _run_gc(BLURAY *bd, gc_ctrl_e msg, uint32_t param)
     int result = -1;
 
     if (bd && bd->graphics_controller && bd->hdmv_vm) {
-        GC_NAV_CMDS cmds = {-1, NULL, -1, 0, 0};
+        GC_NAV_CMDS cmds = {-1, NULL, -1, 0, 0, EMPTY_UO_MASK};
 
         result = gc_run(bd->graphics_controller, msg, param, &cmds);
 
@@ -894,23 +920,6 @@ static int _libbdplus_init(BLURAY *bd)
     return 1;
 }
 
-/*
- * index open
- */
-
-static int _index_open(BLURAY *bd)
-{
-    if (!bd->index) {
-        char *file;
-
-        file = str_printf("%s/BDMV/index.bdmv", bd->device_path);
-        bd->index = indx_parse(file);
-        X_FREE(file);
-    }
-
-    return !!bd->index;
-}
-
 /*
  * meta open
  */
@@ -947,33 +956,82 @@ static void _fill_disc_info(BLURAY *bd)
     bd->disc_info.libjvm_detected = 0;
     bd->disc_info.bdj_handled     = 0;
 
-    if (bd->index) {
+    bd->disc_info.num_titles  = 0;
+    bd->disc_info.titles      = NULL;
+    bd->disc_info.top_menu    = NULL;
+    bd->disc_info.first_play  = NULL;
+
+    array_free((void**)&bd->titles);
+
+    INDX_ROOT *index = indx_parse(bd->device_path);
+    if (index) {
         INDX_PLAY_ITEM *pi;
         unsigned        ii;
 
         bd->disc_info.bluray_detected = 1;
 
-        /* count titles */
+        /* application info */
+        bd->disc_info.video_format     = index->app_info.video_format;
+        bd->disc_info.frame_rate       = index->app_info.frame_rate;
+        bd->disc_info.content_exist_3D = index->app_info.content_exist_flag;
+        bd->disc_info.initial_output_mode_preference = index->app_info.initial_output_mode_preference;
+        memcpy(bd->disc_info.provider_data, index->app_info.user_data, sizeof(bd->disc_info.provider_data));
 
-        for (ii = 0; ii < bd->index->num_titles; ii++) {
-            if (bd->index->titles[ii].object_type == indx_object_type_hdmv) {
+        /* allocate array for title info */
+        BLURAY_TITLE **titles = (BLURAY_TITLE**)array_alloc(index->num_titles + 2, sizeof(BLURAY_TITLE));
+        if (!titles) {
+            BD_DEBUG(DBG_BLURAY | DBG_CRIT, "Can't allocate memory\n");
+            return;
+        }
+        bd->titles = titles;
+        bd->disc_info.titles = (const BLURAY_TITLE * const *)titles;
+        bd->disc_info.num_titles = index->num_titles;
+
+        /* count titles and fill title info */
+
+        for (ii = 0; ii < index->num_titles; ii++) {
+            if (index->titles[ii].object_type == indx_object_type_hdmv) {
                 bd->disc_info.num_hdmv_titles++;
+                titles[ii + 1]->interactive = (index->titles[ii].hdmv.playback_type == indx_hdmv_playback_type_interactive);
+                titles[ii + 1]->id_ref = index->titles[ii].hdmv.id_ref;
             }
-            if (bd->index->titles[ii].object_type == indx_object_type_bdj) {
+            if (index->titles[ii].object_type == indx_object_type_bdj) {
                 bd->disc_info.num_bdj_titles++;
-                bd->disc_info.num_unsupported_titles++;
                 bd->disc_info.bdj_detected = 1;
+                titles[ii + 1]->bdj = 1;
+                titles[ii + 1]->interactive = (index->titles[ii].bdj.playback_type == indx_bdj_playback_type_interactive);
+                titles[ii + 1]->id_ref = atoi(index->titles[ii].bdj.name);
             }
+
+            titles[ii + 1]->accessible =  !(index->titles[ii].access_type & INDX_ACCESS_PROHIBITED_MASK);
+            titles[ii + 1]->hidden     = !!(index->titles[ii].access_type & INDX_ACCESS_HIDDEN_MASK);
         }
 
-        if (bd->index->first_play.object_type == indx_object_type_bdj) {
+        pi = &index->first_play;
+        if (pi->object_type == indx_object_type_bdj) {
             bd->disc_info.bdj_detected = 1;
+            titles[index->num_titles + 1]->bdj = 1;
+            titles[index->num_titles + 1]->interactive = (pi->bdj.playback_type == indx_bdj_playback_type_interactive);
+            titles[index->num_titles + 1]->id_ref = atoi(pi->bdj.name);
         }
-        if (bd->index->top_menu.object_type == indx_object_type_bdj) {
-            bd->disc_info.bdj_detected = 1;
+        if (pi->object_type == indx_object_type_hdmv && pi->hdmv.id_ref != 0xffff) {
+            titles[index->num_titles + 1]->interactive = (pi->hdmv.playback_type == indx_hdmv_playback_type_interactive);
+            titles[index->num_titles + 1]->id_ref = pi->hdmv.id_ref;
         }
 
-        /* BD-J capability */
+        pi = &index->top_menu;
+        if (pi->object_type == indx_object_type_bdj) {
+            bd->disc_info.bdj_detected = 1;
+            titles[0]->bdj = 1;
+            titles[0]->interactive = (pi->bdj.playback_type == indx_bdj_playback_type_interactive);
+            titles[0]->id_ref = atoi(pi->bdj.name);
+        }
+        if (pi->object_type == indx_object_type_hdmv && pi->hdmv.id_ref != 0xffff) {
+            titles[0]->interactive = (pi->hdmv.playback_type == indx_hdmv_playback_type_interactive);
+            titles[0]->id_ref = pi->hdmv.id_ref;
+        }
+
+        /* check for BD-J capability */
 
 #ifdef USING_BDJAVA
         if (bd->disc_info.bdj_detected) {
@@ -987,15 +1045,43 @@ static void _fill_disc_info(BLURAY *bd)
         }
 #endif /* USING_BDJAVA */
 
-        pi = &bd->index->first_play;
+        /* mark supported titles */
+
+        if (bd->disc_info.bdj_detected && !bd->disc_info.bdj_handled) {
+            bd->disc_info.num_unsupported_titles = bd->disc_info.num_bdj_titles;
+        }
+
+        pi = &index->first_play;
         if (pi->object_type == indx_object_type_hdmv && pi->hdmv.id_ref != 0xffff) {
             bd->disc_info.first_play_supported = 1;
         }
+        if (pi->object_type == indx_object_type_bdj) {
+            bd->disc_info.first_play_supported = bd->disc_info.bdj_handled;
+        }
 
-        pi = &bd->index->top_menu;
+        pi = &index->top_menu;
         if (pi->object_type == indx_object_type_hdmv && pi->hdmv.id_ref != 0xffff) {
             bd->disc_info.top_menu_supported = 1;
         }
+        if (pi->object_type == indx_object_type_bdj) {
+            bd->disc_info.top_menu_supported = bd->disc_info.bdj_handled;
+        }
+
+        /* */
+
+        if (bd->disc_info.first_play_supported) {
+            titles[index->num_titles + 1]->accessible = 1;
+            bd->disc_info.first_play = titles[index->num_titles + 1];
+        }
+        if (bd->disc_info.top_menu_supported) {
+            titles[0]->accessible = 1;
+            bd->disc_info.top_menu = titles[0];
+        }
+
+        /* populate title names */
+        bd_get_meta(bd);
+
+        indx_free(&index);
     }
 }
 
@@ -1018,6 +1104,20 @@ uint64_t bd_get_uo_mask(BLURAY *bd)
 
     return mask.u64;
 }
+
+void bd_select_rate(BLURAY *bd, float rate, int reason)
+{
+    if (reason == 2) {
+        /* playback stop. Might want to wait for buffers empty here. */
+        return;
+    }
+    if (rate < 0.5) {
+        _queue_event(bd, BD_EVENT_STILL, 1);
+    } else {
+        _queue_event(bd, BD_EVENT_STILL, 0);
+    }
+}
+
 #endif
 
 #ifdef USING_BDJAVA
@@ -1032,14 +1132,14 @@ uint32_t bd_reg_read(BLURAY *bd, int psr, int reg)
 #endif
 
 #ifdef USING_BDJAVA
-int bd_reg_write(BLURAY *bd, int psr, int reg, uint32_t value)
+int bd_reg_write(BLURAY *bd, int psr, int reg, uint32_t value, uint32_t psr_value_mask)
 {
     if (psr) {
         if (psr < 102) {
             /* avoid deadlocks (psr_write triggers callbacks that may lock this mutex) */
             bd_mutex_lock(&bd->mutex);
         }
-        int res = bd_psr_write(bd->regs, reg, value);
+        int res = bd_psr_write_bits(bd->regs, reg, value, psr_value_mask);
         if (psr < 102) {
             bd_mutex_unlock(&bd->mutex);
         }
@@ -1131,12 +1231,12 @@ static int _start_bdj(BLURAY *bd, unsigned title)
 {
 #ifdef USING_BDJAVA
     if (bd->bdjava == NULL) {
-        bd->bdjava = bdj_open(bd->device_path, bd, bd->index, _bdj_osd_cb, bd->argb_buffer);
+        bd->bdjava = bdj_open(bd->device_path, bd, _bdj_osd_cb, bd->argb_buffer);
         if (!bd->bdjava) {
             return 0;
         }
     }
-    return bdj_start(bd->bdjava, title);
+    return !bdj_process_event(bd->bdjava, BDJ_EVENT_START, title);
 #else
     BD_DEBUG(DBG_BLURAY | DBG_CRIT, "Title %d: BD-J not compiled in\n", title);
     return 0;
@@ -1159,7 +1259,7 @@ static int _bdj_event(BLURAY *bd, unsigned ev, unsigned param)
 static void _stop_bdj(BLURAY *bd)
 {
     if (bd->bdjava != NULL) {
-        bdj_stop(bd->bdjava);
+        bdj_process_event(bd->bdjava, BDJ_EVENT_STOP, 0);
     }
 }
 #else
@@ -1263,6 +1363,13 @@ BLURAY *bd_open(const char* device_path, const char* keyfile_path)
         return NULL;
     }
 
+    bd->regs = bd_registers_init();
+    if (!bd->regs) {
+        BD_DEBUG(DBG_BLURAY, "bd_registers_init() failed\n");
+        X_FREE(bd);
+        return NULL;
+    }
+
     bd->device_path = str_dup(device_path);
 
 #if (defined HAVE_MNTENT_H || defined __APPLE__)
@@ -1271,12 +1378,8 @@ BLURAY *bd_open(const char* device_path, const char* keyfile_path)
 
     _libaacs_init(bd, keyfile_path);
 
-    bd->regs = bd_registers_init();
-
     _libbdplus_init(bd);
 
-    _index_open(bd);
-
     _fill_disc_info(bd);
 
     bd_mutex_init(&bd->mutex);
@@ -1308,13 +1411,13 @@ void bd_close(BLURAY *bd)
     hdmv_vm_free(&bd->hdmv_vm);
 
     gc_free(&bd->graphics_controller);
-    indx_free(&bd->index);
     meta_free(&bd->meta);
     sound_free(&bd->sound_effects);
     bd_registers_free(bd->regs);
 
     _free_event_queue(bd);
     X_FREE(bd->device_path);
+    array_free((void**)&bd->titles);
 
     bd_mutex_destroy(&bd->mutex);
 
@@ -1426,6 +1529,9 @@ int64_t bd_seek_time(BLURAY *bd, uint64_t tick)
         clip = nav_time_search(bd->title, tick, &clip_pkt, &out_pkt);
 
         _seek_internal(bd, clip, out_pkt, clip_pkt);
+
+    } else {
+        BD_DEBUG(DBG_BLURAY | DBG_CRIT, "bd_seek_time(%u) failed\n", (unsigned int)tick);
     }
 
     bd_mutex_unlock(&bd->mutex);
@@ -1443,7 +1549,7 @@ uint64_t bd_tell_time(BLURAY *bd)
     if (bd && bd->title) {
         clip = nav_packet_search(bd->title, SPN(bd->s_pos), &clip_pkt, &out_pkt, &out_time);
         if (clip) {
-            out_time += clip->start_time;
+            out_time += clip->title_time;
         }
     }
 
@@ -1468,6 +1574,9 @@ int64_t bd_seek_chapter(BLURAY *bd, unsigned chapter)
         clip = nav_chapter_search(bd->title, chapter, &clip_pkt, &out_pkt);
 
         _seek_internal(bd, clip, out_pkt, clip_pkt);
+
+    } else {
+        BD_DEBUG(DBG_BLURAY | DBG_CRIT, "bd_seek_chapter(%u) failed\n", chapter);
     }
 
     bd_mutex_unlock(&bd->mutex);
@@ -1524,9 +1633,12 @@ int64_t bd_seek_playitem(BLURAY *bd, unsigned clip_ref)
 
       clip     = &bd->title->clip_list.clip[clip_ref];
       clip_pkt = clip->start_pkt;
-      out_pkt  = clip->pos;
+      out_pkt  = clip->title_pkt;
 
       _seek_internal(bd, clip, out_pkt, clip_pkt);
+
+    } else {
+        BD_DEBUG(DBG_BLURAY | DBG_CRIT, "bd_seek_playitem(%u) failed\n", clip_ref);
     }
 
     bd_mutex_unlock(&bd->mutex);
@@ -1550,6 +1662,9 @@ int64_t bd_seek_mark(BLURAY *bd, unsigned mark)
         clip = nav_mark_search(bd->title, mark, &clip_pkt, &out_pkt);
 
         _seek_internal(bd, clip, out_pkt, clip_pkt);
+
+    } else {
+        BD_DEBUG(DBG_BLURAY | DBG_CRIT, "bd_seek_mark(%u) failed\n", mark);
     }
 
     bd_mutex_unlock(&bd->mutex);
@@ -1653,7 +1768,7 @@ static int _bd_read(BLURAY *bd, unsigned char *buf, int len)
                         if (!_open_m2ts(bd, st)) {
                             return -1;
                         }
-                        bd->s_pos = st->clip->pos;
+                        bd->s_pos = st->clip->title_pkt * 192;
                     } else {
                         _change_angle(bd);
                         _clip_seek_time(bd, bd->angle_change_time);
@@ -1884,7 +1999,8 @@ static int _preload_textst_subpath(BLURAY *bd)
 
 static int _find_ig_stream(BLURAY *bd, uint16_t *pid, int *sub_path_idx, unsigned *sub_clip_idx)
 {
-    MPLS_PI  *pi        = &bd->title->pl->play_item[0];
+    unsigned  main_clip_idx = bd->st0.clip ? bd->st0.clip->ref : 0;
+    MPLS_PI  *pi        = &bd->title->pl->play_item[main_clip_idx];
     unsigned  ig_stream = bd_psr_read(bd->regs, PSR_IG_STREAM_ID);
 
     if (ig_stream > 0 && ig_stream <= pi->stn.num_ig) {
@@ -2020,7 +2136,12 @@ static int _open_playlist(BLURAY *bd, const char *f_name, unsigned angle)
 
     bd_psr_write(bd->regs, PSR_PLAYLIST, atoi(bd->title->name));
     bd_psr_write(bd->regs, PSR_ANGLE_NUMBER, bd->title->angle + 1);
-    bd_psr_write(bd->regs, PSR_CHAPTER, 1);
+
+    if (_is_interactive_title(bd)) {
+        bd_psr_write(bd->regs, PSR_CHAPTER, 0xffff);
+    } else {
+        bd_psr_write(bd->regs, PSR_CHAPTER, 1);
+    }
 
     // Get the initial clip of the playlist
     bd->st0.clip = nav_next_clip(bd->title, NULL);
@@ -2062,6 +2183,44 @@ int bd_select_playlist(BLURAY *bd, uint32_t playlist)
     return result;
 }
 
+#ifdef USING_BDJAVA
+static int _play_playlist_at(BLURAY *bd, int playlist, int playitem, int playmark, int64_t time)
+{
+    if (playlist < 0) {
+        _close_playlist(bd);
+        return 1;
+    }
+
+    if (!bd_select_playlist(bd, playlist)) {
+        return 0;
+    }
+
+    if (playitem > 0) {
+        bd_seek_playitem(bd, playitem);
+    }
+    if (playmark >= 0) {
+        bd_seek_mark(bd, playmark);
+    }
+    if (time >= 0) {
+        bd_seek_time(bd, time);
+    }
+
+    return 1;
+}
+
+int bd_play_playlist_at(BLURAY *bd, int playlist, int playitem, int playmark, int64_t time)
+{
+    int result;
+
+    /* select + seek should be atomic (= player can't read data between select and seek to start position) */
+    bd_mutex_lock(&bd->mutex);
+    result = _play_playlist_at(bd, playlist, playitem, playmark, time);
+    bd_mutex_unlock(&bd->mutex);
+
+    return result;
+}
+#endif /* USING_BDJAVA */
+
 // Select a title for playback
 // The title index is an index into the list
 // established by bd_get_titles()
@@ -2185,7 +2344,7 @@ uint32_t bd_get_titles(BLURAY *bd, uint8_t flags, uint32_t min_title_length)
 
     /* start BD+. No real title info will be passed to BD+ VM ... */
     if (bd->libbdplus) {
-        libbdplus_event(bd->libbdplus, 0xffffffff, bd->index->num_titles, 0);
+        libbdplus_event(bd->libbdplus, 0xffffffff, bd->disc_info.num_titles, 0);
     }
 
     return bd->title_list->count;
@@ -2261,7 +2420,7 @@ static BLURAY_TITLE_INFO* _fill_title_info(NAV_TITLE* title, uint32_t title_idx,
         NAV_CLIP *nc = &title->clip_list.clip[ii];
 
         ci->pkt_count = nc->end_pkt - nc->start_pkt;
-        ci->start_time = (uint64_t)nc->start_time * 2;
+        ci->start_time = (uint64_t)nc->title_time * 2;
         ci->in_time = (uint64_t)pi->in_time * 2;
         ci->out_time = (uint64_t)pi->out_time * 2;
         ci->still_mode = pi->still_mode;
@@ -2381,12 +2540,11 @@ int bd_set_player_setting(BLURAY *bd, uint32_t idx, uint32_t value)
 
     if (idx == BLURAY_PLAYER_SETTING_DECODE_PG) {
         bd_mutex_lock(&bd->mutex);
-        bd->decode_pg = !!value;
 
-        bd_psr_lock(bd->regs);
-        value = (bd_psr_read(bd->regs, PSR_PG_STREAM) & (0x7fffffff)) | ((!!value)<<31);
-        result = !bd_psr_setting_write(bd->regs, PSR_PG_STREAM, value);
-        bd_psr_unlock(bd->regs);
+        bd->decode_pg = !!value;
+        result = bd_psr_write_bits(bd->regs, PSR_PG_STREAM,
+                                   (!!value) << 31,
+                                   0x80000000);
 
         bd_mutex_unlock(&bd->mutex);
         return result;
@@ -2422,15 +2580,13 @@ int bd_set_player_setting_str(BLURAY *bd, uint32_t idx, const char *s)
 
 void bd_select_stream(BLURAY *bd, uint32_t stream_type, uint32_t stream_id, uint32_t enable_flag)
 {
-    uint32_t val;
-
     bd_mutex_lock(&bd->mutex);
-    bd_psr_lock(bd->regs);
 
     switch (stream_type) {
         case BLURAY_PG_TEXTST_STREAM:
-            val = bd_psr_read(bd->regs, PSR_PG_STREAM);
-            bd_psr_write(bd->regs, PSR_PG_STREAM, ((!!enable_flag)<<31) | (stream_id & 0xfff) | (val & 0x7ffff000));
+            bd_psr_write_bits(bd->regs, PSR_PG_STREAM,
+                              ((!!enable_flag)<<31) | (stream_id & 0xfff),
+                              0x80000fff);
             break;
         /*
         case BLURAY_SECONDARY_VIDEO_STREAM:
@@ -2438,7 +2594,6 @@ void bd_select_stream(BLURAY *bd, uint32_t stream_type, uint32_t stream_id, uint
         */
     }
 
-    bd_psr_unlock(bd->regs);
     bd_mutex_unlock(&bd->mutex);
 }
 
@@ -2448,31 +2603,29 @@ void bd_select_stream(BLURAY *bd, uint32_t stream_type, uint32_t stream_id, uint
 
 int bd_start_bdj(BLURAY *bd, const char *start_object)
 {
+    const BLURAY_TITLE *t;
+    int title_num = atoi(start_object);
     unsigned ii;
 
-    if (!bd || !bd->index) {
+    if (!bd) {
         return 0;
     }
 
     /* first play object ? */
-    if (bd->index->first_play.object_type == indx_object_type_bdj &&
-        !strcmp(start_object, bd->index->first_play.bdj.name)) {
-        return _start_bdj(bd, BLURAY_TITLE_FIRST_PLAY);
-    }
-
-    /* top menu ? */
-    if (bd->index->first_play.object_type == indx_object_type_bdj &&
-        !strcmp(start_object, bd->index->top_menu.bdj.name)) {
-        return _start_bdj(bd, BLURAY_TITLE_TOP_MENU);
+    if (bd->disc_info.first_play_supported) {
+        t = bd->disc_info.first_play;
+        if (t && t->bdj && t->id_ref == title_num) {
+            return _start_bdj(bd, BLURAY_TITLE_FIRST_PLAY);
+        }
     }
 
     /* valid BD-J title from disc index ? */
-    for (ii = 0; ii < bd->index->num_titles; ii++) {
-        INDX_TITLE *t = &bd->index->titles[ii];
-
-        if (t->object_type == indx_object_type_bdj &&
-            !strcmp(start_object, t->bdj.name)) {
-            return _start_bdj(bd, ii + 1);
+    if (bd->disc_info.titles) {
+        for (ii = 0; ii <= bd->disc_info.num_titles; ii++) {
+            t = bd->disc_info.titles[ii];
+            if (t && t->bdj && t->id_ref == title_num) {
+                return _start_bdj(bd, ii);
+            }
         }
     }
 
@@ -2557,6 +2710,7 @@ static void _process_psr_write_event(BLURAY *bd, BD_PSR_EVENT *ev)
             libbdplus_event(bd->libbdplus, 0x110, ev->new_val, 0);
             break;
         case PSR_PLAYLIST:
+            _bdj_event  (bd, BDJ_EVENT_PLAYLIST,ev->new_val);
             _queue_event(bd, BD_EVENT_PLAYLIST, ev->new_val);
             break;
         case PSR_PLAYITEM:
@@ -2608,11 +2762,13 @@ static void _process_psr_change_event(BLURAY *bd, BD_PSR_EVENT *ev)
             }
 
             bd_mutex_lock(&bd->mutex);
+            if (bd->st0.clip) {
             _init_pg_stream(bd);
             if (bd->st_textst.clip) {
                 BD_DEBUG(DBG_BLURAY | DBG_CRIT, "Changing TextST stream\n");
                 _preload_textst_subpath(bd);
             }
+            }
             bd_mutex_unlock(&bd->mutex);
 
             break;
@@ -2717,7 +2873,8 @@ static int _play_hdmv(BLURAY *bd, unsigned id_ref)
     bd->title_type = title_hdmv;
 
     if (!bd->hdmv_vm) {
-        bd->hdmv_vm = hdmv_vm_init(bd->device_path, bd->regs, bd->index);
+        bd->hdmv_vm = hdmv_vm_init(bd->device_path, bd->regs, bd->disc_info.num_titles,
+                                   bd->disc_info.first_play_supported, bd->disc_info.top_menu_supported);
     }
 
     if (hdmv_vm_select_object(bd->hdmv_vm, id_ref)) {
@@ -2738,24 +2895,20 @@ static int _play_title(BLURAY *bd, unsigned title)
 {
     /* first play object ? */
     if (title == BLURAY_TITLE_FIRST_PLAY) {
-        INDX_PLAY_ITEM *p = &bd->index->first_play;
 
         bd_psr_write(bd->regs, PSR_TITLE_NUMBER, 0xffff); /* 5.2.3.3 */
 
-        if (p->object_type == indx_object_type_hdmv) {
-            if (p->hdmv.id_ref == 0xffff) {
-                /* no first play title (5.2.3.3) */
-                bd->title_type = title_hdmv;
-                return 1;
-            }
-            return _play_hdmv(bd, p->hdmv.id_ref);
+        if (!bd->disc_info.first_play_supported) {
+            /* no first play title (5.2.3.3) */
+            bd->title_type = title_hdmv;
+            return 1;
         }
 
-        if (p->object_type == indx_object_type_bdj) {
+        if (bd->disc_info.first_play->bdj) {
             return _play_bdj(bd, title);
+        } else {
+            return _play_hdmv(bd, bd->disc_info.first_play->id_ref);
         }
-
-        return 0;
     }
 
     /* bd_play not called ? */
@@ -2766,36 +2919,32 @@ static int _play_title(BLURAY *bd, unsigned title)
 
     /* top menu ? */
     if (title == BLURAY_TITLE_TOP_MENU) {
-        INDX_PLAY_ITEM *p = &bd->index->top_menu;
 
         bd_psr_write(bd->regs, PSR_TITLE_NUMBER, 0); /* 5.2.3.3 */
 
-        if (p->object_type == indx_object_type_hdmv) {
-            if (p->hdmv.id_ref == 0xffff) {
-                /* no top menu (5.2.3.3) */
-                bd->title_type = title_hdmv;
-                return 0;
-            }
-            return _play_hdmv(bd, p->hdmv.id_ref);
+        if (!bd->disc_info.top_menu_supported) {
+            /* no top menu (5.2.3.3) */
+            bd->title_type = title_hdmv;
+            return 0;
         }
 
-        if (p->object_type == indx_object_type_bdj) {
+        if (bd->disc_info.top_menu->bdj) {
             return _play_bdj(bd, title);
+        } else {
+            return _play_hdmv(bd, bd->disc_info.top_menu->id_ref);
         }
 
         return 0;
     }
 
     /* valid title from disc index ? */
-    if (title > 0 && title <= bd->index->num_titles) {
-        INDX_TITLE *t = &bd->index->titles[title-1];
+    if (title > 0 && title <= bd->disc_info.num_titles) {
 
         bd_psr_write(bd->regs, PSR_TITLE_NUMBER, title); /* 5.2.3.3 */
-
-        if (t->object_type == indx_object_type_hdmv) {
-            return _play_hdmv(bd, t->hdmv.id_ref);
-        } else {
+        if (bd->disc_info.titles[title]->bdj) {
             return _play_bdj(bd, title);
+        } else {
+            return _play_hdmv(bd, bd->disc_info.titles[title]->id_ref);
         }
     }
 
@@ -3053,7 +3202,7 @@ static int _read_ext(BLURAY *bd, unsigned char *buf, int len, BD_EVENT *event)
     }
 
     if (bd->bdj_end_of_playlist == 1) {
-        _bdj_event(bd, BDJ_EVENT_END_OF_PLAYLIST, 0);
+        _bdj_event(bd, BDJ_EVENT_END_OF_PLAYLIST, bd_psr_read(bd->regs, PSR_PLAYLIST));
         bd->bdj_end_of_playlist |= 2;
     }
 
@@ -3236,6 +3385,8 @@ int bd_get_sound_effect(BLURAY *bd, unsigned sound_id, BLURAY_SOUND_EFFECT *effe
 
 const struct meta_dl *bd_get_meta(BLURAY *bd)
 {
+    const struct meta_dl *meta = NULL;
+
     if (!bd) {
         return NULL;
     }
@@ -3248,11 +3399,22 @@ const struct meta_dl *bd_get_meta(BLURAY *bd)
 
     if (psr_menu_lang != 0 && psr_menu_lang != 0xffffff) {
         const char language_code[] = {(psr_menu_lang >> 16) & 0xff, (psr_menu_lang >> 8) & 0xff, psr_menu_lang & 0xff, 0 };
-        return meta_get(bd->meta, language_code);
+        meta = meta_get(bd->meta, language_code);
+    } else {
+        meta = meta_get(bd->meta, NULL);
     }
-    else {
-        return meta_get(bd->meta, NULL);
+
+    /* assign title names to disc_info */
+    if (meta) {
+        unsigned ii;
+        for (ii = 0; ii < meta->toc_count; ii++) {
+            if (meta->toc_entries[ii].title_number > 0 && meta->toc_entries[ii].title_number <= bd->disc_info.num_titles) {
+                bd->titles[meta->toc_entries[ii].title_number]->name = meta->toc_entries[ii].title_name;
+            }
+        }
     }
+
+    return meta;
 }
 
 struct clpi_cl *bd_get_clpi(BLURAY *bd, unsigned clip_ref)
@@ -3273,3 +3435,13 @@ void bd_free_clpi(struct clpi_cl *cl)
 {
     clpi_free(cl);
 }
+
+struct mpls_pl *bd_read_mpls(const char *mpls_file)
+{
+    return mpls_parse(mpls_file);
+}
+
+void bd_free_mpls(struct mpls_pl *pl)
+{
+    mpls_free(pl);
+}
