@@ -73,6 +73,7 @@ struct graphics_controller_s {
     unsigned        auto_action_triggered;
     BOG_DATA       *bog_data;
     BOG_DATA       *saved_bog_data;
+    BD_UO_MASK      page_uo_mask;
 
     /* page effects */
     int                    effect_idx;
@@ -367,6 +368,10 @@ static int _save_page_state(GRAPHICS_CONTROLLER *gc)
 
     X_FREE(gc->saved_bog_data);
     gc->saved_bog_data = calloc(page->num_bogs, sizeof(*gc->saved_bog_data));
+    if (!gc->saved_bog_data) {
+        GC_ERROR("_save_page_state(): out of memory\n");
+        return -1;
+    }
 
     for (ii = 0; ii < page->num_bogs; ii++) {
         gc->saved_bog_data[ii].enabled_button = gc->bog_data[ii].enabled_button;
@@ -682,6 +687,7 @@ static void _select_page(GRAPHICS_CONTROLLER *gc, uint16_t page_id, int out_effe
     _select_button(gc, button_id);
 
     gc->valid_mouse_position = 0;
+    gc->page_uo_mask = bd_empty_uo_mask();
 
     if (out_effects) {
         page = _find_page(&gc->igs->ics->interactive_composition, cur_page_id);
@@ -713,6 +719,7 @@ static void _gc_reset(GRAPHICS_CONTROLLER *gc)
 
     gc->popup_visible = 0;
     gc->valid_mouse_position = 0;
+    gc->page_uo_mask = bd_empty_uo_mask();
 
     graphics_processor_free(&gc->igp);
     graphics_processor_free(&gc->pgp);
@@ -774,6 +781,10 @@ static void _process_psr_event(void *handle, BD_PSR_EVENT *ev)
 GRAPHICS_CONTROLLER *gc_init(BD_REGISTERS *regs, void *handle, gc_overlay_proc_f func)
 {
     GRAPHICS_CONTROLLER *p = calloc(1, sizeof(*p));
+    if (!p) {
+        GC_ERROR("gc_init(): out of memory\n");
+        return NULL;
+    }
 
     p->regs = regs;
 
@@ -825,6 +836,9 @@ int gc_decode_ts(GRAPHICS_CONTROLLER *gc, uint16_t pid, uint8_t *block, unsigned
 
         if (!gc->igp) {
             gc->igp = graphics_processor_init();
+            if (!gc->igp) {
+                return -1;
+            }
         }
 
         bd_mutex_lock(&gc->mutex);
@@ -865,6 +879,9 @@ int gc_decode_ts(GRAPHICS_CONTROLLER *gc, uint16_t pid, uint8_t *block, unsigned
         /* PG stream */
         if (!gc->pgp) {
             gc->pgp = graphics_processor_init();
+            if (!gc->pgp) {
+                return -1;
+            }
         }
         graphics_processor_decode_ts(gc->pgp, &gc->pgs,
                                      pid, block, num_blocks,
@@ -881,6 +898,9 @@ int gc_decode_ts(GRAPHICS_CONTROLLER *gc, uint16_t pid, uint8_t *block, unsigned
         /* TextST stream */
         if (!gc->tgp) {
             gc->tgp = graphics_processor_init();
+            if (!gc->tgp) {
+                return -1;
+            }
         }
         graphics_processor_decode_ts(gc->tgp, &gc->tgs,
                                      pid, block, num_blocks,
@@ -921,6 +941,9 @@ int gc_add_font(GRAPHICS_CONTROLLER *p, const char *font_file)
 
     if (!p->textst_render) {
         p->textst_render = textst_render_init();
+        if (!p->textst_render) {
+            return -1;
+        }
     }
 
     return textst_render_add_font(p->textst_render, font_file);
@@ -1042,15 +1065,19 @@ static int _render_textst(GRAPHICS_CONTROLLER *p, uint32_t stc, GC_NAV_CMDS *cmd
                 continue;
             }
 
-            TEXTST_BITMAP bmp = {NULL, style->text_box.width, style->text_box.height, style->text_box.width};
+            TEXTST_BITMAP bmp = {NULL, style->text_box.width, style->text_box.height, style->text_box.width, 0};
             bmp.mem = malloc(bmp.width * bmp.height);
-            memset(bmp.mem, style->region_info.background_color, bmp.width * bmp.height);
+            if (bmp.mem) {
+                memset(bmp.mem, style->region_info.background_color, bmp.width * bmp.height);
 
-            textst_render(p->textst_render, &bmp, style, region);
+                textst_render(p->textst_render, &bmp, style, region);
 
-            _render_textst_region(p, dialog[ii].start_pts, style, &bmp, s->style->palette);
+                _render_textst_region(p, dialog[ii].start_pts, style, &bmp, s->style->palette);
 
-            X_FREE(bmp.mem);
+                X_FREE(bmp.mem);
+            } else {
+                GC_ERROR("_render_textst(): out of memory\n");
+            }
         }
 
         /* commit changes */
@@ -1377,6 +1404,7 @@ static int _render_page(GRAPHICS_CONTROLLER *gc,
                   s->ics->video_descriptor.video_height);
     }
 
+    gc->page_uo_mask = page->uo_mask_table;
 
     for (ii = 0; ii < page->num_bogs; ii++) {
         BD_IG_BOG    *bog      = &page->bog[ii];
@@ -1874,6 +1902,7 @@ int gc_run(GRAPHICS_CONTROLLER *gc, gc_ctrl_e ctrl, uint32_t param, GC_NAV_CMDS
         cmds->nav_cmds     = NULL;
         cmds->sound_id_ref = -1;
         cmds->status       = GC_STATUS_NONE;
+        cmds->page_uo_mask = bd_empty_uo_mask();
     }
 
     if (!gc) {
@@ -2009,6 +2038,10 @@ int gc_run(GRAPHICS_CONTROLLER *gc, gc_ctrl_e ctrl, uint32_t param, GC_NAV_CMDS
                 cmds->status |= GC_STATUS_MENU_OPEN;
             }
         }
+
+        if (gc->ig_open && !gc->out_effects) {
+            cmds->page_uo_mask = gc->page_uo_mask;
+        }
     }
 
     bd_mutex_unlock(&gc->mutex);
